/* Quaternion representation of rotations. */

/* See, e.g.: http://www.cs.berkeley.edu/~laura/cs184/quat/quaternion.html. */

#include <stdio.h>
#include <math.h>
#include "util.h"

/* Given an axis and an angle, compute a rotation quaternion. */

void axis_to_quaternion(float *axis, float theta, QUAT *quat)
{
	quat->s = cos(theta / 2.);
	vcopy(axis, quat->v);
	vnormalize(quat->v);
	vscale(quat->v, sin(theta / 2.));
}

/* Given two rotations, q1 and q2, expressed as quaternions,
quaternion multiplication yields the equivalent product rotation.
This routine also normalizes the result every COUNT times it is
called, to keep error from creeping in. */

#define COUNT 100

void qmult(QUAT *q1, QUAT *q2, QUAT *dest)
{
	XYZ v1, v2;
	QUAT q;
	static int count = 0;

	vcopy(q1->v, v1);
	vcopy(q2->v, v2);
	vscale(v1, q2->s);
	vscale(v2, q1->s);
	vcross(q1->v, q2->v, q.v);
	vadd(v1, q.v, q.v);
	vadd(v2, q.v, q.v);
	q.s = q1->s * q2->s - vdot(q1->v, q2->v);

	*dest = q;

	if (++count > COUNT) {
		count = 0;
#ifdef notdef
		debug("length of q is %g\n",
			sqrt(q.s * q.s + q.v[0] * q.v[0] + q.v[1] * q.v[1] +
			q.v[2] * q.v[2]));
#endif
		qnormalize(dest);
	}
}

/* Normalize greatest component, to avoid problems that occur when the
component we're normalizing gets close to zero (and the other components may
add up to more than 1.0 because of rounding error).  */

void qnormalize(QUAT *q)
{
	int which, i;
	float gr, v[4];

	vcopy(q->v, v);
	v[3] = q->s;

	which = 0;
	gr = v[0];
	for (i = 1; i < 4; i++) {
		if (fabs(v[i]) > fabs(gr)) {
			gr = v[i];
			which = i;
		}
	}

	v[which] = 0.;
	v[which] = sqrt(1.0 - (v[0] * v[0] + v[1] * v[1] +
		v[2] * v[2] + v[3] * v[3]));

	/* Check to see if we need negative square root. */

	if (gr < 0.0) {
		v[which] = -v[which];
	}
	vcopy(v, q->v);
	q->s = v[3];
}

/* Build a rotation matrix from a rotation quaternion. */

void quaternion_to_rotmatrix(QUAT *q, float *m)
{
	float w, x, y, z;

	w = q->s;
	x = q->v[0];
	y = q->v[1];
	z = q->v[2];

	/* Column major order, used to left multiply points, i.e., P'=M*P */

	m[0]  = 1. - 2. * (y * y + z * z);
	m[1]  = 2. * (x * y + w * z);
	m[2]  = 2. * (x * z - w * y);
	m[3]  = 0.;

	m[4]  = 2. * (x * y - w * z);
	m[5]  = 1. - 2. * (x * x + z * z);
	m[6]  = 2. * (y * z + w * x);
	m[7]  = 0.;

	m[8]  = 2. * (x * z + w * y);
	m[9]  = 2. * (y * z - w * x);
	m[10] = 1. - 2. * (x * x + y * y);
	m[11] = 0.;

	m[12] = 0.;
	m[13] = 0.;
	m[14] = 0.;
	m[15] = 1.;
}
