/* Generic OpenGL/glut user interface for 3D display of objects that uses a
trackball for positioning.  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/glut.h>
#include "util.h"

static QUAT Spin;               /* quaternion incremental rotation */
static QUAT Rot;                /* quaternion total rotation */
static XYZ Trans;               /* total translation */
static XYZ OTrans;              /* object total translation */

static float Mscale;            /* mouse movement multiplier */

static int Mbut;                /* mouse button status */
static int Mod;                 /* modifier keys */
static int Mx;                  /* place where mouse is now */
static int My;
static int Oldx;                /* place where mouse used to be */
static int Oldy;
static int Width;               /* size of window */
static int Height;

#define MLEFT 1                 /* Mbut bits */
#define MMIDDLE 2
#define MRIGHT 4

#ifdef EMULATE3BUTTON
#define PAN (MLEFT)
#define ZOOM (MMIDDLE)
#define SPIN (MRIGHT)
#else
#define PAN (MLEFT)
#define ZOOM (MLEFT | MMIDDLE)
#define SPIN (MMIDDLE)
#endif

static void trackball(QUAT *q, int oldx, int oldy, int newx, int newy);

static void Idle(void)
{
	float *v, dx, dy;
	XYZ trans;

	/* Shift- PAN or ZOOM translates the unrotated scene. */

	v = (Mod & GLUT_ACTIVE_SHIFT) ? OTrans : Trans;

	/* If the user drags the mouse, Spin or [O]Trans are updated.  If no
	mouse buttons are pressed, just keep on applying the previous spin
	rotation, over and over.  */

	if (Mbut) {
		dx = (float)(Mx - Oldx) * Mscale / (float)Width;
		dy = (float)(Oldy - My) * Mscale / (float)Height;
		if (Mbut == PAN) {
			vset(trans, dx, dy, 0.);
			vadd(trans, v, v);
		} else if (Mbut == ZOOM) {
			vset(trans, 0., 0., dx + dy);
			vadd(trans, v, v);
		} else if (Mbut == SPIN) {

			/* Shift SPIN only uses the larger component of
			the motion. */

			if (Mod & GLUT_ACTIVE_SHIFT) {
				if (fabs(dx) > fabs(dy)) {
					Oldy = My;
				} else {
					Oldx = Mx;
				}
			}

			trackball(&Spin, Oldx, Oldy, Mx, My);
			qmult(&Spin, &Rot, &Rot);
		}
		Oldx = Mx;
		Oldy = My;
	} else {
		qmult(&Spin, &Rot, &Rot);
	}
	glutPostRedisplay();
}

/* Click. */

static void Mouse(int button, int state, int mx, int my)
{
	/* Get window state and modifiers. */

	Width = glutGet(GLUT_WINDOW_WIDTH);
	Height = glutGet(GLUT_WINDOW_HEIGHT);
	Mod = glutGetModifiers();

	/* Record mouse state. */

	if (state == GLUT_DOWN) {
		Oldx = Mx = mx;
		Oldy = My = my;

		switch (button) {
		case GLUT_LEFT_BUTTON:
			Mbut |= MLEFT;
			break;

		case GLUT_MIDDLE_BUTTON:
			Mbut |= MMIDDLE;
			break;

		case GLUT_RIGHT_BUTTON:
			Mbut |= MRIGHT;
			break;
		}
	} else {
		switch (button) {
		case GLUT_LEFT_BUTTON:
			Mbut &= ~MLEFT;
			break;

		case GLUT_MIDDLE_BUTTON:
			Mbut &= ~MMIDDLE;
			break;

		case GLUT_RIGHT_BUTTON:
			Mbut &= ~MRIGHT;
			break;
		}
	}
}

/* Drag. */

static void MouseMotion(int mx, int my)
{
	Mx = mx;
	My = my;
}

/* Initialize the trackball interface.  Call after other glut initialization. */

void tbInit(float scale)
{
	/* Callbacks used by the trackball. */

	glutIdleFunc(Idle);
	glutMouseFunc(Mouse);
	glutMotionFunc(MouseMotion);

	/* Default is no rotation or translation. */

	Spin.s = 1.0;
	vzero(Spin.v);
	Rot.s = 1.0;
	vzero(Rot.v);
	vzero(Trans);
	vzero(OTrans);

	Mscale = scale;

	/* Initialize interface variables.  The others are set the
	first time the user clicks the mouse. */

	Mbut = 0;
	Mod = 0;
}

/* Save and restore the trackball state. */

void tbSave(char *name)
{
	FILE *outfile;

	outfile = fileopen(name, "w");
	fprintf(outfile, "%g %g %g %g\n", Rot.s, Rot.v[0], Rot.v[1], Rot.v[2]);
	fprintf(outfile, "%g %g %g\n", Trans[0], Trans[1], Trans[2]);
	fprintf(outfile, "%g %g %g\n", OTrans[0], OTrans[1], OTrans[2]);
	fclose(outfile);
}

void tbRestore(char *name)
{
	char buf[100];
	FILE *infile;

	infile = fileopen(name, "r");
	fgets(buf, 100, infile);
	sscanf(buf, "%g %g %g %g", &Rot.s, &Rot.v[0], &Rot.v[1], &Rot.v[2]);
	fgets(buf, 100, infile);
	sscanf(buf, "%g %g %g", &Trans[0], &Trans[1], &Trans[2]);
	fgets(buf, 100, infile);
	sscanf(buf, "%g %g %g", &OTrans[0], &OTrans[1], &OTrans[2]);
	fclose(infile);
}

/* Call this in the display callback just before rendering the scene.
This performs the translations and rotation specified by the trackball. */

void tbTransformScene(int origin)
{
	float m[16];

	/* Translate (rotated coordinates). */

	glTranslatef(Trans[0], Trans[1], Trans[2]);

	/* Rotate. */

	quaternion_to_rotmatrix(&Rot, m);
	glMultMatrixf(m);

	if (origin) {
		glutSolidSphere(1., 10, 10);
	}

	/* Translate (unrotated coordinates). */

	glTranslatef(OTrans[0], OTrans[1], OTrans[2]);
}

/* Set the trackball's center of rotation. */

void tbOrigin(float x, float y, float z)
{
	vset(OTrans, x, y, z);
}

/* Pan or zoom. */

void tbTranslate(float *v)
{
	vadd(Trans, v, Trans);
}

/* Rotate the trackball (and the scene) theta degrees around an axis. */

void tbRotate(float *axis, float theta)
{
	QUAT q;

	axis_to_quaternion(axis, theta * M_PI / 180., &q);
	qmult(&q, &Rot, &Rot);
}

/* Implementation of a virtual trackball.  Original code by Gavin Bell, lots
of ideas from Thant Tessman and the August '88 issue of SigGraph's "Computer
Graphics," pp. 121-129. */

/* Size of virtual trackball, in relation to window size. */

#define TRACKBALLSIZE 1.0

static void window_to_tb(float mx, float my, float *x, float *y);
static float tb_project_to_sphere(float, float, float);

/* Ok, simulate a trackball.  Project the mouse positions onto the virtual
trackball, then figure out the axis of rotation, which is the cross product
of O P1 and O P2 (O is the center of the ball).  Note: This is a deformed
trackball -- it's a trackball in the center, but is deformed into a
hyperbolic solid of rotation away from the center.  */

static void trackball(QUAT *q, int oldx, int oldy, int newx, int newy)
{
	float p1x, p1y, p2x, p2y;
	float theta;
	XYZ p1, p2, axis, d;

	if (newx == oldx && newy == oldy) {

		/* Zero rotation. */

		q->s = 1.;
		vzero(q->v);
		return;
	}

	/* Scale the old and new mouse positions to (-1, 1). */

	window_to_tb(oldx, oldy, &p1x, &p1y);
	window_to_tb(newx, newy, &p2x, &p2y);

	/* First, figure out z-coordinates for projection of P1 and P2 to
	deformed sphere. */

	vset(p1, p1x, p1y, tb_project_to_sphere(TRACKBALLSIZE, p1x, p1y));
	vset(p2, p2x, p2y, tb_project_to_sphere(TRACKBALLSIZE, p2x, p2y));

	/* Now the axis of rotation. */

	vcross(p1, p2, axis);

	/* Figure out how much to rotate around that axis. */

	vsub(p1, p2, d);
	theta = sin(vlength(d) * Mscale / 5. / (2.0 * TRACKBALLSIZE));

	/* Return a rotation quaternion. */

	axis_to_quaternion(axis, theta, q);
}

/* Map mouse click mx, my to a more convenient (-1.0, 1.0) range, based on
window size.  */

static void window_to_tb(float mx, float my, float *x, float *y)
{
	*x = (2.0 * mx) / (float)Width - 1.;
	*y = 1. - (2.0 * my) / (float)Height;
}

/* Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet if we
are away from the center of the sphere.  */

static float tb_project_to_sphere(float r, float x, float y)
{
	float d, t, z;

	d = sqrt(x * x + y * y);
	if (d < r * M_SQRT1_2) {        /* inside sphere */
		z = sqrt(r * r - d * d);
	} else {                        /* on hyperbola */
		t = r / M_SQRT2;
		z = t * t / d;
	}
	return (z);
}
